
## Automatic trace method switching

Particularly away from the MLE, the symmetrized stochastic trace estimators are
nearly exact to single precision (see [1] in the README for figures and
demonstrations). And they are of course *much* faster to compute than the exact
traces. So it can in some cases be very worthwhile to do the first handful of
optimization steps using the stochastic traces and then switch to exact when the
optimizer is warmed up enough.

There are two main options to do this, and both are set with global variables.
The first option is to switch from stochastic traces to exact traces when the
$\ell_{\infty}$ norm of the object goes below a certain number. So, for example:
* `GPMaxlik.SETTINGS.GRAD_INFNORM_TOL = 5.0` means that when the inf norm of the
  gradient goes below five, every subsequent call to `gnll`, even when `saa !=
  nothing`, will compute the trace for the gradient exactly.
* `GPMaxlik.SETTINGS.FISH_INFNORM_TOL = ...` means the same for the expected
  Fisher matrix.

The second option is just to set `GPMaxlik.SETTINGS.STOCH_MAXIT=100`, for
example, to switch to exact traces after `100` calls to `gnll`, in whatever form.

To reset this counter and any other manually changed settings to the defaults,
simply call `GPMaxlik.reset_settings!()`.

If you want to completely disable stochastic trace estimation for one of the
options but always use it for the other, you also have the variables
`GPMaxlik.SETTINGS.STOCH_GRAD` and `GPMaxlik.SETTINGS.STOCH_FISH`, which can be
set to `true` or `false`. These mean what they look like, and setting these to
`false` overrides any other setting and will lead to exact traces being used.

Some general guidelines:
* You really can benefit from at least doing the first few iterations with
  stochastic traces. It is *so* much faster.
* Near the MLE where the gradient is near zero, the stochastic trace in the
  gradient can start to make trouble. So it is advisable to switch to exact
  gradient traces for high-tolerance optimization. BUT, **the stochastic
  expected Fisher is more or less always accurate to nearly single precision**.
  So you probably never need to exactly compute that quantity.

