---
toc: false
---

The main function, `gnll`, has the following arguments: 
```julia
gnll(
# positional arguments:
pts,
data,
covfn,
derivs_covfn,
params;
# keyword arguments:
saa,
nll,
grad,
fish,
pscale=false,
profile=false,
vrb=false,
K=nothing,
Kdv=DerivativeCovarianceMatrix[],
store=nothing
)
```
Where:
* `pts` is some iterable list of locations at which the measurements have been
made. Typically a `Vector{SVector{D,Float64}}` for a `D`-dimensional process.

* `data` is a `Vector{Float64}` or `Vector{Vector{Float64}}` representing your
measurements.

* `covfn` is your covariance function. This object does **not** need to be a
subtype of function (as in, `covfn <: Function` does not need to return
`true`). But it has to have the method `covfn(x1, x2, params)`, where `x1` and
`x2` are members of `pts`.

* `derivs_covfn` is some iterable list of the derivatives of your covariance
function with respect to each of the parameters, each function with the same
signature as `covfn`. That is, `derivs_covfn[j](x1, x2, params)` gives back
`(d/dparams_j)covfn(x1, x2, params)|_{params_j}`.

* `params` is some iterable collection of parameters.

* `saa` is the argument that determines whether you use stochastic trace
estimation to approximate the trace terms in the gradient and expected Fisher
matrix. If `saa=nothing`, then the gradient and/or expected fisher matrix will
be computed exactly. If it is a `Matrix{Float64}` of dimension (`n` x `M`),
where `n` is the length of your data and `M` is some typically small number,
then each column of that matrix will be treated as an SAA vector and stochastic
trace approximation will be used instead of exact traces, *so long as the global
stochastic flags also permit it*. See the section on automatic trace method
switching for more information.

* `nll`, `grad`, `fish`, and `pscale` are `Bool` options that indicate whether
you want those quantities to be computed. Here `pscale` is the implied scale
parameter from the profile likelihood.

* `vrb` is the verbose flag.

* `K` is **any** object that implements the methods required for covariance
  matrices (see [the section on covariance operators](./docs/operators.md)).

* `Kdv` is some iterable list of operators for the derivatives of the covariance
matrices with respect to kernel parameters. Again, see [the section on
covariance operators](./docs/operators.md)

* `store` is the optional object that stores intermediate results, primarily
useful when using an optimizer that asks for gradients and Hessians via
different functions (like Ipopt). For more details on that, see [the section on
storage](./docs/storage.md).

