---
toc: false
---

## Storage objects

A significant advantage of using the symmetrized Hutchinson trace estimators for
the traces in the gradient and expected Fisher matrix is that computing the
gradient means that you get the expected Fisher almost for free. Put
differently, the call `gnll(..., nll=true, grad=true, fish=false)` is not going
to be much faster than `gnll(..., nll=true, grad=true, fish=true)`. So with that
in mind, it would be nice to save the intermediate quantities that make this
fact true. Similarly, calling `gnll(..., nll=true, grad=false, fish=true)` and
then calling `gnll(..., nll=true, grad=true, fish=false)` afterwards will again
make significant redundant work.

Unfortunately, lots of optimization software ask for gradient and Hessian
functions to be provided separately (looking at you, Ipopt). In order to still
enjoy the benefit of sensibly re-using the redundant objects in computing the
stochastic gradient and expected Fisher matrix, `GPMaxlik` allows you to provide
a `store` object `S`, again a potentially arbitrary `struct`, that implements the
following methods, where `key::UInt64`:
* `Base.delete!(S, key)`
* `Base.getindex(S, key)`
* `Base.setindex!(S, key)`
* `Base.haskey(S, key)`

These names are sufficiently self-explanatory that I'll spare you the
exposition. The astute reader will also notice that these are just the standard
methods for working with `Dict` structures, and you're exactly right. You could
provide a raw `Dict{UInt64, T} where{T}` here.

See [this example
implementation](https://git.sr.ht/~cgeoga/GPMaxlik.jl/tree/master/item/example/example_store.jl)
that actually writes the intermediate files to disk in a `tmp` file, which is
very nice to have so you don't eat any un-necessary RAM. This store is used in
[this example
file](https://git.sr.ht/~cgeoga/GPMaxlik.jl/tree/master/item/example/example_ipopt.jl)
which optimizes using Ipopt. You can try using and not using the store to see
the speed gains. For many parameters, they are substantial! If nothing else, it
saves a whole pass of calling `update!` on derivative covariance operators.

