
## A simple trust region optimizer

Primarily for convenience, we also provide a simple trust region optimizer based
off of the textbook chapter in Nocedal and Wright (2006). Its signature is given
by
```julia
GPMaxlik.trustregion(
# positional arguments:
f,    # objective function
fgh,  # a 3-tuple of (objective, gradient, Hessian)
init; # initial point
# keyword arguments:
vrb=false,  # verbose informative printing
dmax=1.0,   # maximum radius of trust region
dini=0.5,   # initial radius of trust region
eta=0.125,  # tuning parameter, see Nocedal and Wright
rtol=1e-8,  # stopping rtol condition (in objective)
atol=1e-5,  # stopping atol condition (in objective)
gtol=1e-4,  # stopping gradient norm (or Newton step) condition 
maxit=200,  # maximum iterations
dcut=1e-4,  # cutoff and return if trust region radius gets below this value
# advanced usage kwargs (see below):
iter_funs=(convg_info, check_obj_nan),
iter_funs_kwargs=tuple()
)
```

Since most of these arguments are self-explanatory we will only discuss and add
a few additional points:
* The output of `fgh` should be a 3-tuple. It should either be `(obj, grad,
  Hessian)` or `(obj, grad, nothing)`, where in the latter case BFGS updates to
  a Hessian approximation will be used.
* After updating the gradient and Hessian, arbitrary functions provided by
  `iter_funs` will be run that have a specific signature. This is an opportunity
for your to add your own specific print statements, stopping conditions, and the
like. This is actually how the convergence checker is implemented---see
[here](https://git.sr.ht/~cgeoga/GPMaxlik.jl/tree/master/item/src/trustregion.jl#L111)
for a demonstration.

In general, it is advisable for you to use a more sophisticated optimizer. But
this optimizer we provide here does have a few nice properties that make it
useful in some cases:
* Sophisticated tools like Ipopt expect exact gradients and Hessians, so once
  you get near the MLE you might bounce around forever and never "converge".
* The calling syntax here makes it as easy as possible to do no redundant work
  with respect to sequential gradient and Hessian computations.
* The arbitrary `iter_funs` loop every iteration gives you the opportunity to
  really watch the progress of the optimizer and print diagnostics. Like, if you
want to [watch the condition of your
Hessian](https://git.sr.ht/~cgeoga/GPMaxlik.jl/tree/master/item/src/trustregion.jl#L152),
that is just a few extra lines in this setting.

We've used this in several big projects and gotten what we needed out of it.
Outside of maximum likelhood for GPs we make no promises of it being at all
useful. But for GPs its track record is pretty good for settings where it is
applicable.
