
using Test
include("../example/simulate_data.jl")

# Just for one data point.
function naive_nll(p)
  K = cholesky!(Symmetric([kfn(x,y,p) for x in pts, y in pts]))
  0.5*(logdet(K) + dot(simul, K\simul))
end

# Basic expected fisher.
function efish(p)
  out = zeros(2,2)
  K   = cholesky!(Symmetric([kfn(x,y,p) for x in pts, y in pts]))
  Kd1 = Symmetric([kd1(x,y,p) for x in pts, y in pts])
  Kd2 = Symmetric([kd2(x,y,p) for x in pts, y in pts])
  out[1,1] = tr(K\(Kd1*(K\Kd1)))
  out[1,2] = tr(K\(Kd2*(K\Kd1)))
  out[2,1] = out[1,2]
  out[2,2] = tr(K\(Kd2*(K\Kd2)))
  Symmetric(out./2)
end

@testset "exact gnll core functions" begin
  GPMAXTEST = GPMaxlik.gnll(pts, simul, kfn, (kd1, kd2), [3.0, 3.0], saa=nothing,
                            nll=true, grad=true, fish=true, vrb=false)
  @test isapprox(GPMAXTEST.nll, naive_nll([3.0, 3.0]))
  @test isapprox(GPMAXTEST.grad, ForwardDiff.gradient(naive_nll, [3.0, 3.0]))
  @test isapprox(GPMAXTEST.fish, efish([3.0, 3.0]))
end

@testset "stochastic gnll core functions" begin
  GPMAXTEST_SAA = GPMaxlik.gnll(pts, simul, kfn, (kd1, kd2), [3.0, 3.0], saa=saav,
                                nll=true, grad=true, fish=true, vrb=false)
  @test isapprox(GPMAXTEST_SAA.nll, naive_nll([3.0, 3.0]))
  stoch_g = GPMAXTEST_SAA.grad
  true_g  = ForwardDiff.gradient(naive_nll, [3.0, 3.0])
  @test isapprox(stoch_g, true_g, rtol=1e-3)
  stoch_f = GPMAXTEST_SAA.fish
  true_f  = efish([3.0, 3.0])
  @test isapprox(stoch_f, true_f, rtol=1e-3)
end

@testset "manual AD rules" begin
  _nll_fwd = p -> GPMaxlik.gnll_forwarddiff(p, pts, simul, kfn)
  @test isapprox(_nll_fwd([3.0, 3.0]), naive_nll([3.0, 3.0]))
  @test isapprox(ForwardDiff.gradient(_nll_fwd,  [3.0, 3.0]), 
                 ForwardDiff.gradient(naive_nll, [3.0, 3.0]))
  @test isapprox(ForwardDiff.hessian(_nll_fwd,   [3.0, 3.0]), 
                 ForwardDiff.hessian(naive_nll,  [3.0, 3.0]))
end


